Step: 1
Step: 2
Step: 3
Step: 4
Step: 5
Step: 6
go(red,east,1) go(red,east,3) go(red,north,4) go(red,south,2) go(red,south,6) go(red,west,5) position(blue,1,16,0) position(blue,1,16,1) position(blue,1,16,2) position(blue,1,16,3) position(blue,1,16,4) position(blue,1,16,5) position(blue,1,16,6) position(green,16,1,0) position(green,16,1,1) position(green,16,1,2) position(green,16,1,3) position(green,16,1,4) position(green,16,1,5) position(green,16,1,6) position(red,1,1,0) position(red,2,1,1) position(red,2,2,2) position(red,3,1,5) position(red,3,15,6) position(red,4,1,4) position(red,4,2,3) position(yellow,16,16,0) position(yellow,16,16,1) position(yellow,16,16,2) position(yellow,16,16,3) position(yellow,16,16,4) position(yellow,16,16,5) position(yellow,16,16,6)
SAT
