Step: 1
domain(p(1),true,false) domain(p(2),false,true) domain(p(3),false,false) domain(q(1),true,false) domain(q(2),false,false) domain(q(3),false,false) domain_of_p(p(1),true,false) domain_of_p(p(2),false,true) domain_of_p(p(3),false,false) in_domain(p(2),true) in_domain(p(4),false) sig(p,1) sig(q,1) universe(6)
UNKNOWN
