//
// Copyright (c) 2015-2017 Benjamin Kaufmann
//
// This file is part of Potassco.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// deal in the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.
#include <potassco/clingo.h>
namespace Potassco {
AbstractAssignment::~AbstractAssignment() {}
bool AbstractAssignment::isTotal()          const { return unassigned() == 0u; }
bool AbstractAssignment::isFixed(Lit_t lit) const { return value(lit) != Value_t::Free && level(lit) == 0; }
bool AbstractAssignment::isTrue(Lit_t lit)  const { return value(lit) == Value_t::True; }
bool AbstractAssignment::isFalse(Lit_t lit) const { return value(lit) == Value_t::False; }
uint32_t AbstractAssignment::trailEnd(uint32_t level) const {
	return level < this->level() ? this->trailBegin(level + 1) : this->trailSize();
}

AbstractSolver::~AbstractSolver() {}
AbstractPropagator::~AbstractPropagator() {}
AbstractHeuristic::~AbstractHeuristic() {}
AbstractStatistics::~AbstractStatistics() {}

} // namespace Potassco
