Step: 1
pattern(1,1)
pattern(1,2)
pattern(1,3)
pattern(1,5)
Step: 2
pattern(1,1) pattern(2,1)
pattern(1,1) pattern(2,2)
pattern(1,2) pattern(2,1)
pattern(1,2) pattern(2,2)
pattern(1,2) pattern(2,5)
pattern(1,5) pattern(2,2)
pattern(1,5) pattern(2,5)
Step: 3
pattern(1,2) pattern(2,1) pattern(3,2)
pattern(1,2) pattern(2,2) pattern(3,1)
pattern(1,2) pattern(2,2) pattern(3,2)
pattern(1,2) pattern(2,5) pattern(3,2)
pattern(1,5) pattern(2,2) pattern(3,2)
pattern(1,5) pattern(2,2) pattern(3,5)
pattern(1,5) pattern(2,5) pattern(3,2)
Step: 4
pattern(1,2) pattern(2,2) pattern(3,1) pattern(4,2)
pattern(1,2) pattern(2,2) pattern(3,2) pattern(4,1)
pattern(1,2) pattern(2,2) pattern(3,2) pattern(4,2)
pattern(1,5) pattern(2,2) pattern(3,2) pattern(4,2)
pattern(1,5) pattern(2,2) pattern(3,5) pattern(4,2)
SAT
