// {{{ MIT License

// Copyright 2017 Roland Kaminski

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// deal in the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:

// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.

// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.

// }}}

#include "tests/tests.hh"
#include "gringo/hash_set.hh"

namespace Gringo { namespace Test {

TEST_CASE("unique_vec", "[base]") {
    SECTION("push") {
        UniqueVec<unsigned> vec;
        REQUIRE( vec.push(1).second);
        REQUIRE( vec.push(2).second);
        REQUIRE( vec.push(3).second);
        REQUIRE( vec.push(4).second);
        vec.pop();
        vec.pop();
        REQUIRE(!vec.push(1).second);
        REQUIRE(!vec.push(2).second);
        REQUIRE( vec.push(3).second);
        REQUIRE( vec.push(4).second);
        REQUIRE( vec.push(5).second);
        REQUIRE( vec.push(6).second);
        REQUIRE( vec.push(7).second);
        vec.erase([](unsigned val) { return 2 < val && val < 6; });
        REQUIRE(!vec.push(1).second);
        REQUIRE(!vec.push(2).second);
        REQUIRE( vec.push(3).second);
        REQUIRE( vec.push(4).second);
        REQUIRE( vec.push(5).second);
        REQUIRE(!vec.push(6).second);
        REQUIRE(!vec.push(7).second);
    }
}

} } // namespace Test Gringo


