Step: 1
Step: 2
Step: 3
occurs(cpa_start(cpa_n1),1) occurs(cpa_travel(cpa_n1,cpa_n2),2)
occurs(cpa_start(cpa_n2),1) occurs(cpa_travel(cpa_n2,cpa_n1),2)
Step: 4
Step: 5
Step: 6
occurs(cpa_start(cpa_n1),1) occurs(cpa_start(cpa_n2),2) occurs(cpa_travel(cpa_n1,cpa_n2),3) occurs(cpa_travel(cpa_n2,cpa_n1),4)
occurs(cpa_start(cpa_n1),1) occurs(cpa_start(cpa_n2),2) occurs(cpa_travel(cpa_n1,cpa_n2),3) occurs(cpa_travel(cpa_n2,cpa_n1),4)
occurs(cpa_start(cpa_n1),1) occurs(cpa_start(cpa_n2),2) occurs(cpa_travel(cpa_n1,cpa_n2),4) occurs(cpa_travel(cpa_n2,cpa_n1),3)
occurs(cpa_start(cpa_n1),1) occurs(cpa_start(cpa_n2),2) occurs(cpa_travel(cpa_n1,cpa_n2),4) occurs(cpa_travel(cpa_n2,cpa_n1),3)
occurs(cpa_start(cpa_n1),2) occurs(cpa_start(cpa_n2),1) occurs(cpa_travel(cpa_n1,cpa_n2),3) occurs(cpa_travel(cpa_n2,cpa_n1),4)
occurs(cpa_start(cpa_n1),2) occurs(cpa_start(cpa_n2),1) occurs(cpa_travel(cpa_n1,cpa_n2),3) occurs(cpa_travel(cpa_n2,cpa_n1),4)
occurs(cpa_start(cpa_n1),2) occurs(cpa_start(cpa_n2),1) occurs(cpa_travel(cpa_n1,cpa_n2),4) occurs(cpa_travel(cpa_n2,cpa_n1),3)
occurs(cpa_start(cpa_n1),2) occurs(cpa_start(cpa_n2),1) occurs(cpa_travel(cpa_n1,cpa_n2),4) occurs(cpa_travel(cpa_n2,cpa_n1),3)
SAT
