Step: 1
a c
c
Step: 2
a c res(atoms(b,c)) res(atoms(c)) res(model("a","c")) res(model("c")) res(shown(a,c)) res(shown(c)) res(terms(a)) res(terms)
c res(atoms(b,c)) res(atoms(c)) res(model("a","c")) res(model("c")) res(shown(a,c)) res(shown(c)) res(terms(a)) res(terms)
SAT
